;---------------------------------------------------------------------------
;
; 		       32-bit PM Kernel Memory Manager
;
;     Copyright (c) 1999-2001 by Thomas Kjoernes, all rights reserved.
;
;---------------------------------------------------------------------------

	INCLUDE	std.inc
	INCLUDE	stdgroup.inc
	INCLUDE	kernel.api

	.386p

	; Exports

	PUBLIC	kmmHeapAllocate
	PUBLIC	kmmHeapFree

	; Startup/Shutdown

	PUBLIC	kmmInit
	PUBLIC	kmmDone

codeSegment SEGMENT

;---------------------------------------------------------------------------
;	kmmGetPDE
;
;	Entry:
;		eax = Linear Address
;	Exit:
;		esi = Linear Address of PDE for [eax]
;
;	The active PD is always mapped starting at 0xFFFFF000
;
DefProc kmmGetPDE

	mov	esi, eax
	and	esi, 0FFC00000h
	shr	esi, 20
	or	esi, 0FFFFF000h
	ret

EndProc kmmGetPDE

;---------------------------------------------------------------------------
;	kmmGetPTE
;
;	Entry:
;		eax = Linear Address
;	Exit:
;		esi = Linear Address of PTE for [eax]
;
;	The active PTs are always mapped starting at 0xFFC00000
;
DefProc kmmGetPTE

	mov	esi, eax
	and	esi, 0FFFFF000h
	shr	esi, 10
	or	esi, 0FFC00000h
	ret

EndProc kmmGetPTE

;---------------------------------------------------------------------------
;	kmmGetMasterPDE
;
;	Entry:
;		eax = Linear Address
;	Exit:
;		esi = Linear Address of master PDE for [eax]
;
;	The master PD is always mapped at 0xFFBFF000 and it is used to keep
;	track of shared PTs.
;
;	Whenever a PDE for a shared PT is updated, for instance when adding, 
;	removing or when changing the address, the master PD is updated along
;	with the active PD. 
;
;	The PD for all other address spaces is not updated until there is a 
;	page fault and mismatch is found between the then active PD and the 
;	master PD.
;
DefProc kmmGetMasterPDE

	mov	esi, eax
	and	esi, 0FFC00000h
	shr	esi, 20
	or	esi, 0FFBFF000h
	ret

EndProc kmmGetMasterPDE

;---------------------------------------------------------------------------
;	kmmPPageAllocate
;
;	Entry:
;		ecx = Size
;
;	Exit (Success CF==0):
;
;		eax = Page Address 'Physical'
;
;	Exit (Failure CF==1):
;
;		Allocation Failure
;
;	kmmPPageAllocate allocates a contigous range of physical pages
;	from the list of free physical memory ranges.
;
;
DefProc kmmPPageAllocate

	ret

EndProc kmmPPageAllocate

;---------------------------------------------------------------------------
;	kmmPPageFree
;
;	Entry:
;		ebx = Page Address 'Physical'
;		ecx = Size
;
;	Exit:
;		CF==0 if Successful
;
DefProc kmmPPageFree

	ret

EndProc kmmPPageFree

;---------------------------------------------------------------------------
;	kmmPPageRegister
;
;	Entry:
;		ebx = Block Base Address 'Physical'
;		ecx = Block Size
;	Exit:
;		None
;
DefProc kmmPPageRegister

	push	eax
	push	ebx
	push	ecx
	push	edx

	mov	eax, ebx

	add	eax, 000000FFFh
	and	eax, 0FFFFF000h

	mov	edx, eax
	sub	eax, ebx

	sub	ecx, eax
	and	ecx, 0FFFFF000h
	jz	kmmPPageRegisterExit

	push	ecx
	mov	ecx, SIZE LIST_HEAD + 8
	APIcall	KERNEL_HeapAllocate
	pop	ecx
	jc	kmmPPageRegisterExit

	lea	ebx, kmmFreeList
	APIcall	KERNEL_ListAttachTail

	mov	[eax+08h], edx
	mov	[eax+0Ch], ecx

kmmPPageRegisterExit:
	pop	edx
	pop	ecx
	pop	ebx
	pop	eax
	ret

EndProc kmmPPageRegister

;---------------------------------------------------------------------------
;	kmmPPageGetFreeList
;
;	Entry:
;		None
;	Exit:
;		ebx = handle of kmmFreeList
;
DefProc kmmPPageGetFreeList

	mov	ebx, OFFSET kmmFreeList
	ret

EndProc kmmPPageGetFreeList

;---------------------------------------------------------------------------
;			Kernel Heap Allocation
;---------------------------------------------------------------------------

HEAP_USED	EQU	BYTE PTR 1
HEAP_LAST	EQU	BYTE PTR 2

;---------------------------------------------------------------------------
;	kmmHeapAllocate - Allocate Kernel Heap memory
;
;	Entry:
;		ecx = Size
;
;	Exit (Success CF==0):
;
;		eax = Block Address 'Linear'
;
;	Exit (Failure CF==1):
;
;		Allocation Failure
;
DefProc kmmHeapAllocate

	push	ecx
	push	esi	

	mov	esi, kmmHeapBase		; esi -> Kernel Heap

	add	ecx, 3
	and	ecx, 0FFFFFFFCh

kmmHeapAllocateFind:
	mov	eax, [esi]		;
	and	eax, 0FFFFFFFCh		;

	test	[esi], HEAP_USED	; used block?
	jz	kmmHeapAllocateSize	;

	test	[esi], HEAP_LAST	; last block?
	jz	kmmHeapAllocateSkip	;

	stc
	jmp	kmmHeapAllocateExit

kmmHeapAllocateSkip:
	lea	esi, [esi+eax+4]	; esi -> next block
	jmp	kmmHeapAllocateFind	;

kmmHeapAllocateSize:
	cmp	eax, ecx		; large enough for request?
	jb	kmmHeapAllocateSkip	; smaller?
	jz	kmmHeapAllocateDone	; perfect?

	sub	eax, ecx

	; split block...

	sub	eax, 4			; reserve size-tag space
	je	kmmHeapAllocateDone	; worth it?

	test	[esi], HEAP_LAST	; is this the last block?

	mov	[esi], eax		; save 'free block' size
	lea	esi, [esi+eax+4]	; esi -> next block
	mov	[esi], ecx		; save 'used block' size

	jz	kmmHeapAllocateDone	; last?

	or	[esi], HEAP_LAST	; mark block as 'Last'

kmmHeapAllocateDone:
	or	[esi], HEAP_USED	; mark block as 'Used'
	lea	eax, [esi+4]		; eax -> block
	clc

kmmHeapAllocateExit:
	pop	esi
	pop	ecx
	ret

EndProc kmmHeapAllocate

;---------------------------------------------------------------------------
;	kmmHeapFree - Free Kernel Heap memory
;
;	Entry:
;		ebx = Block Address 'Linear'
;	Exit:
;		CF==0 if Successful
;
DefProc kmmHeapFree

	push	eax
	push	ebx
	push	esi

	mov	esi, kmmHeapBase
	sub	ebx, 4

kmmHeapFreeFind:
	cmp	ebx, esi
	je	kmmHeapFreeMain

	test	[esi], HEAP_LAST	; last block?
	jnz	kmmHeapFreeFail		;

	mov	eax, [esi]		; eax = size
	and	eax, 0FFFFFFFCh		;

	lea	esi, [esi+eax+4]	; esi -> next block
	jmp	kmmHeapFreeFind		;

kmmHeapFreeMain:
	test	[esi], HEAP_USED	; used block?
	jz	kmmHeapFreeFail		;

	xor	[esi], HEAP_USED
	call	kmmHeapOptimize

kmmHeapFreeDone:
	clc

kmmHeapFreeExit:
	pop	esi
	pop	ebx
	pop	eax
	ret

kmmHeapFreeFail:
	stc
	jmp	kmmHeapFreeExit

EndProc kmmHeapFree

;---------------------------------------------------------------------------
;	kmmHeapOptimize - Optimize Heap by merging free blocks
;
;	Entry:
;		None
;	Exit:
;		None
;
DefProc kmmHeapOptimize

	push	eax
	push	ebx
	push	edx
	push	esi

	mov	esi, kmmHeapBase

kmmHeapOptimizeFind:
	mov	eax, [esi]		; eax = size
	and	eax, 0FFFFFFFCh		;

	test	[esi], HEAP_USED	; used block?
	jz	kmmHeapOptimizeMain	;

kmmHeapOptimizeSkip:
	test	[esi], HEAP_LAST	; last block?
	jnz	kmmHeapOptimizeDone	;

	lea	esi, [esi+eax+4]	; esi -> next block
	jmp	kmmHeapOptimizeFind	;

kmmHeapOptimizeMain:
	mov	ebx, esi		; ebx -> block
	lea	esi, [esi+eax+4]	; esi -> next block

kmmHeapOptimizeNext:
	test	[esi], HEAP_USED	; used block?
	jnz	kmmHeapOptimizeFind	;

	mov	edx, [esi]
	and	edx, 0FFFFFFFCh

	lea	eax, [eax+edx+4]

	mov	[ebx], eax

	test	[esi], HEAP_LAST	; last block?
	jnz	kmmHeapOptimizeLast	;

	lea	esi, [ebx+eax+4]	; esi -> next block
	jmp	kmmHeapOptimizeNext	;

kmmHeapOptimizeLast:
	or	[ebx], HEAP_LAST

kmmHeapOptimizeDone:
	pop	esi
	pop	edx
	pop	ebx
	pop	eax
	ret

EndProc kmmHeapOptimize

;---------------------------------------------------------------------------
;			Startup and Shutdown
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
;	kmmInit
;
;	Entry (Registers):
;
;		ebp = pointer to arguments on the stack
;
;	Entry (Stack):
;
;		DWORD = return address
;
;	Entry (Stack Arguments):
;
;		DWORD	Base of Library
;		DWORD	Size of Library
;		DWORD	Base of Reclaim Area
;		DWORD	Size of Reclaim Area
;		DWORD	Base of MDT
;		DWORD	Size of MDT (number of entries)
;		DWORD	Base of PSP
;		DWORD	Base of DOS Environment
;
;	Exit:
;	     	None
;
DefProc kmmInit

	; initialize Kernel Heap

	mov	DWORD PTR kmmHeapPool[0], SIZE kmmHeapPool-4 OR HEAP_LAST

	; register physical pages

	mov	esi, [ebp].kaMDTBase
	mov	eax, [ebp].kaMDTSize

kmmInitNext:
	mov	ebx, [esi].mdtBase
	mov	ecx, [esi].mdtSize
	call	kmmPPageRegister

kmmInitSkip:
	add	esi, SIZE MDT
	dec	eax
	jnz	kmmInitNext

	; map master page directory at 0xFFBFF000

;	mov	eax, cr3
;	mov	edx, 0FFBFF000h
;	call	kmmMapLinearToPhysical

	ret

EndProc kmmInit

;---------------------------------------------------------------------------
;	kmmDone
;
;	Entry:
;		Done
;	Exit:
;	     	None
;
DefProc kmmDone
	ret
EndProc kmmDone

codeSegment ENDS

;---------------------------------------------------------------------------
; Data
;---------------------------------------------------------------------------

dataSegment SEGMENT

kmmHeapBase	DD	kmmHeapPool

dataSegment ENDS

;---------------------------------------------------------------------------
; BSS Data
;---------------------------------------------------------------------------

bssSegment SEGMENT

kmmHeapPool	DB	_1MEG / 4 DUP (?)

kmmFreeList	LIST_HEAD <>

bssSegment ENDS

	END

